# @Collection.register
# class MSMarco(Collection):
#     module_name = "msmarco"
#     config_keys_not_in_path = ["path"]
#     collection_type = "TrecCollection"
#     generator_type = "DefaultLuceneDocumentGenerator"
#     config_spec = [ConfigOption("path", "/GW/NeuralIR/nobackup/msmarco/trec_format", "path to corpus")]
